import os
import tarfile
import urllib.request

wayland_version = "1.21.0"
protocols_version = "1.25"

wayland_source = "https://cgit.freedesktop.org/wayland/wayland/plain/protocol/wayland.xml?id={}".format(
    wayland_version
)
protocols_source = (
    "https://wayland.freedesktop.org/releases/wayland-protocols-{}.tar.xz".format(
        protocols_version
    )
)


def _is_within_directory(directory, target):
    """Helper to check for CVE-2007-4559"""
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


def _safe_extractall(tar, path=".", members=None, *, numeric_owner=False):
    """Helper to check for CVE-2007-4559"""
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not _is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(path, members, numeric_owner=numeric_owner)


def protocols_build(output_dir):
    from pywayland.scanner import Protocol

    # first, we download the wayland.xml file
    wayland_file = "wayland.xml"
    urllib.request.urlretrieve(wayland_source, wayland_file)

    # download the protocols file and extract it
    protocol_dest = f"wayland-protocols-{protocols_version}"
    urllib.request.urlretrieve(protocols_source, protocol_dest + ".tar.xz")

    with tarfile.open(protocol_dest + ".tar.xz") as f:
        _safe_extractall(f)

    # walk the directory and generate all the protocols
    protocol_files = [
        wayland_file,
        *sorted(
            [
                os.path.join(dirpath, filename)
                for dirpath, _, filenames in os.walk(protocol_dest)
                for filename in filenames
                if os.path.splitext(filename)[1] == ".xml"
            ],
            reverse=True,
        ),
    ]

    protocols = [Protocol.parse_file(protocol_file) for protocol_file in protocol_files]
    protocol_imports = {
        interface.name: protocol.name
        for protocol in protocols
        for interface in protocol.interface
    }
    for protocol in protocols:
        protocol.output(output_dir, protocol_imports)



#!/usr/bin/env python
#
# Copyright (c) 2024 -- Lars Heuer
# All rights reserved.
#
# SPDX-License-Identifier: CC-PDDC
#
"""\
Prints information about a Wayland compositor.
"""
from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field
from functools import partial
from itertools import chain
from operator import itemgetter

from pywayland.client import Display
from pywayland.protocol.wayland import WlOutput, WlRegistry, WlSeat, WlShm


@dataclass
class Info:
    common: list = field(default_factory=list)
    seat: list = field(default_factory=list)
    output: list = field(default_factory=list)
    shm: list = field(default_factory=list)
    roundtrip_needed: bool = False

    def __str__(self) -> str:

        def stringify_interface(interface: str, version: int, name: int) -> str:
            return f"{interface:<{pad}}version: {version:2}, name: {name:2}"

        def stringify_head_tail(l: list) -> Iterator[str]:  # noqa: E741
            head, *tail = l
            yield stringify_interface(*head)
            yield from tail

        pad = len(max(map(itemgetter(0), self.common), key=len))

        return "\n".join(
            chain(
                stringify_head_tail(self.seat),
                stringify_head_tail(self.shm),
                stringify_head_tail(self.output),
                (
                    stringify_interface(*e)
                    for e in sorted(self.common, key=itemgetter(0))
                ),
            )
        )


def _add_interface_info(
    l: list, *, interface: str, version: int, name: int
) -> None:  # noqa: E741
    # Not building the complete output string here because
    # we need the length of the 1st element, see pad in Info.__str__()
    l.append((f"interface: '{interface}', ", version, name))


def add_seat_info(
    info: Info, registry: WlRegistry, id_num: int, interface: str, version: int
) -> None:

    def handle_capabilities(_, capabilities):
        caps = sorted(cap.name for cap in WlSeat.capability if capabilities & cap.value)
        append(f"capabilities: {' '.join(caps)}")

    def handle_name(_, name):
        append(f"name: {name}")

    _add_interface_info(info.seat, interface=interface, version=version, name=id_num)
    append = lambda s: info.seat.append(f"    {s}")  # noqa: E731
    seat = registry.bind(id_num, WlSeat, version)
    seat.dispatcher["name"] = handle_name
    seat.dispatcher["capabilities"] = handle_capabilities


def add_output_info(
    info: Info, registry: WlRegistry, id_num: int, interface: str, version: int
) -> None:

    def handle_geometry(
        _,
        x: int,
        y: int,
        width: int,
        height: int,
        subpixel: int,
        make: str,
        model: str,
        transform: int,
    ) -> None:
        subpixel_info = next(m.name for m in WlOutput.subpixel if m.value == subpixel)
        transform_info = next(
            m.name for m in WlOutput.transform if m.value == transform
        )
        append(f"x: {x}, y: {y}")
        append(f"physical_width: {width} mm, physical_height: {height} mm")
        append(f"make: '{make}', model: '{model}'")
        append(
            f"subpixel_orientation: {subpixel_info}, output_transform: {transform_info}"
        )

    def handle_scale(_, scale: int):
        append(f"scale: {scale}")

    def handle_name(_, name):
        append(f"name: {name}")

    def handle_description(_, description):
        append(f"description: {description}")

    def handle_mode(_, flags: int, width: int, height: int, refresh: int) -> None:
        flag_info = next(m.name for m in WlOutput.mode if flags & m.value)
        append("mode:")
        append_sub = lambda s: info.output.append(f"        {s}")  # noqa: E731
        append_sub(
            f"width: {width} px, height: {height} px, refresh: {refresh / 1000} Hz"
        )
        append_sub(f"flags: {flag_info}")

    _add_interface_info(info.output, interface=interface, version=version, name=id_num)
    append = lambda s: info.output.append(f"    {s}")  # noqa: E731
    output = registry.bind(id_num, WlOutput, version)
    output.dispatcher["name"] = handle_name
    output.dispatcher["description"] = handle_description
    output.dispatcher["geometry"] = handle_geometry
    output.dispatcher["scale"] = handle_scale
    output.dispatcher["mode"] = handle_mode


def add_shm_info(
    info: Info, registry: WlRegistry, id_num: int, interface: str, version: int
) -> None:

    def handle_format(_, fmt: int) -> None:
        format_info = next(f for f in WlShm.format if f.value == fmt)
        append(f"{hex(format_info.value)}: {format_info.name}")

    _add_interface_info(info.shm, interface=interface, version=version, name=id_num)
    append = lambda s: info.shm.append(f"    {s}")  # noqa: E731
    append("formats:")
    shm = registry.bind(id_num, WlShm, version)
    shm.dispatcher["format"] = handle_format


def handle_registry_global(
    info: Info, registry: WlRegistry, id_num: int, interface: str, version: int
) -> None:
    if interface in ("wl_seat", "wl_output", "wl_shm"):
        globals()[f"add_{interface[3:]}_info"](
            info, registry, id_num, interface, version
        )
        info.roundtrip_needed = True
    else:
        _add_interface_info(
            info.common, interface=interface, version=version, name=id_num
        )


def main():
    with Display() as display:
        registry = display.get_registry()
        info = Info()
        registry.dispatcher["global"] = partial(handle_registry_global, info)
        display.dispatch()
        while True:
            info.roundtrip_needed = False
            display.roundtrip()
            if not info.roundtrip_needed:
                break
        print(info)


if __name__ == "__main__":
    main()



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.statemachine import ViewList
from jinja2 import Template
from sphinx.util.docstrings import prepare_docstring
from sphinx.util.nodes import nested_parse_with_titles


# Adapted from sphinxcontrib-httpdomain
def import_object(module_name, class_name):
    mod = importlib.import_module(module_name)
    return getattr(mod, class_name)


def format_args(func):
    sig = inspect.signature(func)
    parameters = [v for k, v in sig.parameters.items() if k != "self"]
    sig = sig.replace(parameters=parameters)
    return str(sig)


wl_protocol_template = Template(
    """
.. autoclass:: {{ module }}.{{ class_name }}
    {% for func, opcode, sig, docs in requests %}
    .. method:: {{ func }} {{ sig }}

        .. rubric:: Request -- opcode {{ opcode}} (attached to :class:`~pywayland.server.resource.Resource` instance)

{% for doc in docs %}
        {{ doc }}{% endfor %}{% endfor %}
    {% for func, opcode, sig, docs in events %}
    .. method:: {{ func }} {{ sig }}

        .. rubric:: Event -- opcode {{ opcode}} (attached to :class:`~pywayland.client.proxy.Proxy` instance)

{% for doc in docs %}
        {{ doc }}{% endfor %}{% endfor %}
"""
)


class WlProtocol(Directive):
    required_arguments = 2

    def make_rst(self):
        module_name, class_name = self.arguments[:2]
        obj = import_object(module_name, class_name)
        events = [
            (
                event.py_func.__name__,
                opcode,
                format_args(event.py_func),
                prepare_docstring(event.py_func.__doc__),
            )
            for opcode, event in enumerate(obj.events)
        ]
        reqs = [
            (
                req.py_func.__name__,
                opcode,
                format_args(req.py_func),
                prepare_docstring(req.py_func.__doc__),
            )
            for opcode, req in enumerate(obj.requests)
        ]
        context = {
            "module": module_name,
            "class_name": class_name,
            "obj": obj,
            "events": events,
            "requests": reqs,
        }
        rst = wl_protocol_template.render(**context)
        yield from rst.splitlines()

    def run(self):
        node = nodes.section()
        node.document = self.state.document
        result = ViewList()
        for line in self.make_rst():
            result.append(line, f"<{self.__class__.__name__}>")
        nested_parse_with_titles(self.state, result, node)
        return node.children


def setup(app):
    app.add_directive("wl_protocol", WlProtocol)
